@@ -4,6 +4,7 @@ This script tests :class:`.PathSpec`.
 
 import os
 import os.path
+import pathlib
 import shutil
 import tempfile
 import unittest
@@ -14,6 +15,11 @@ from pathspec import (
 	PathSpec)
 from pathspec.util import (
 	iter_tree_entries)
+from pathspec.patterns.gitwildmatch import GitWildMatchPatternError
+from tests.util import (
+	make_dirs,
+	make_files,
+	ospath)
 
 
 class PathSpecTest(unittest.TestCase):
@@ -25,36 +31,19 @@ class PathSpecTest(unittest.TestCase):
 		"""
 		Create the specified directories.
 		"""
-		for dir in dirs:
-			os.mkdir(os.path.join(self.temp_dir, self.ospath(dir)))
+		make_dirs(self.temp_dir, dirs)
 
 	def make_files(self, files: Iterable[str]) -> None:
 		"""
 		Create the specified files.
 		"""
-		for file in files:
-			self.mkfile(os.path.join(self.temp_dir, self.ospath(file)))
-
-	@staticmethod
-	def mkfile(file: str) -> None:
-		"""
-		Creates an empty file.
-		"""
-		with open(file, 'wb'):
-			pass
-
-	@staticmethod
-	def ospath(path: str) -> str:
-		"""
-		Convert the POSIX path to a native OS path.
-		"""
-		return os.path.join(*path.split('/'))
+		return make_files(self.temp_dir, files)
 
 	def setUp(self) -> None:
 		"""
 		Called before each test.
 		"""
-		self.temp_dir = tempfile.mkdtemp()
+		self.temp_dir = pathlib.Path(tempfile.mkdtemp())
 
 	def tearDown(self) -> None:
 		"""
@@ -131,6 +120,32 @@ class PathSpecTest(unittest.TestCase):
 			'./src/test2/c/c.txt',
 		})
 
+	def test_01_empty_path(self):
+		"""
+		Tests that patterns that end with an escaped space will be treated properly.
+		"""
+		spec = PathSpec.from_lines('gitwildmatch', [
+			'\\ ',
+			'abc\\ '
+		])
+		test_files = [
+			' ',
+			'  ',
+			'abc ',
+			'somefile',
+		]
+		results = list(filter(spec.match_file, test_files))
+		self.assertEqual(results, [
+			' ',
+			'abc '
+		])
+
+		# An escape with double spaces is invalid.
+		# Disallow it. Better to be safe than sorry.
+		self.assertRaises(GitWildMatchPatternError, lambda: PathSpec.from_lines('gitwildmatch', [
+			'\\  '
+		]))
+
 	def test_01_match_files(self):
 		"""
 		Tests that matching files one at a time yields the same results as
@@ -311,12 +326,12 @@ class PathSpecTest(unittest.TestCase):
 			__entry.path
 			for __entry in spec.match_entries(entries)
 		}
-		self.assertEqual(results, {
+		self.assertEqual(results, set(map(ospath, [
 			'X/a.txt',
 			'X/Z/c.txt',
 			'Y/a.txt',
 			'Y/Z/c.txt',
-		})
+		])))
 
 	def test_05_match_file(self):
 		"""
@@ -390,12 +405,12 @@ class PathSpecTest(unittest.TestCase):
 			__entry.path
 			for __entry in spec.match_tree_entries(self.temp_dir)
 		}
-		self.assertEqual(results, {
+		self.assertEqual(results, set(map(ospath, [
 			'X/a.txt',
 			'X/Z/c.txt',
 			'Y/a.txt',
 			'Y/Z/c.txt',
-		})
+		])))
 
 	def test_05_match_tree_files(self):
 		"""
@@ -420,12 +435,12 @@ class PathSpecTest(unittest.TestCase):
 			'Y/Z/c.txt',
 		])
 		results = set(spec.match_tree_files(self.temp_dir))
-		self.assertEqual(results, {
+		self.assertEqual(results, set(map(ospath, [
 			'X/a.txt',
 			'X/Z/c.txt',
 			'Y/a.txt',
 			'Y/Z/c.txt',
-		})
+		])))
 
 	def test_06_issue_41_a(self):
 		"""
@@ -537,3 +552,80 @@ class PathSpecTest(unittest.TestCase):
 		self.assertEqual(results, {
 			'anydir/file.txt',
 		})
+
+	def test_08_issue_39(self):
+		"""
+		Test excluding files in a directory.
+		"""
+		spec = PathSpec.from_lines('gitwildmatch', [
+			'*.log',
+			'!important/*.log',
+			'trace.*',
+		])
+		files = {
+			'a.log',
+			'b.txt',
+			'important/d.log',
+			'important/e.txt',
+			'trace.c',
+		}
+		ignores = set(spec.match_files(files))
+		self.assertEqual(ignores, {
+			'a.log',
+			'trace.c',
+		})
+		self.assertEqual(files - ignores, {
+			'b.txt',
+			'important/d.log',
+			'important/e.txt',
+		})
+
+	def test_09_issue_80_a(self):
+		"""
+		Test negating patterns.
+		"""
+		spec = PathSpec.from_lines('gitwildmatch', [
+			'build',
+			'*.log',
+			'.*',
+			'!.gitignore',
+		])
+		files = {
+			'.c-tmp',
+			'.gitignore',
+			'a.log',
+			'b.txt',
+			'build/d.log',
+			'build/trace.bin',
+			'trace.c',
+		}
+		keeps = set(spec.match_files(files, negate=True))
+		self.assertEqual(keeps, {
+			'.gitignore',
+			'b.txt',
+			'trace.c',
+		})
+
+	def test_09_issue_80_b(self):
+		"""
+		Test negating patterns.
+		"""
+		spec = PathSpec.from_lines('gitwildmatch', [
+			'build',
+			'*.log',
+			'.*',
+			'!.gitignore',
+		])
+		files = {
+			'.c-tmp',
+			'.gitignore',
+			'a.log',
+			'b.txt',
+			'build/d.log',
+			'build/trace.bin',
+			'trace.c',
+		}
+		keeps = set(spec.match_files(files, negate=True))
+		ignores = set(spec.match_files(files))
+		self.assertEqual(files - ignores, keeps)
+		self.assertEqual(files - keeps, ignores)
